@@ -24,10 +24,8 @@
 
 #define MAXNAMES 200
 
-static int help      = 0;
 static int relative  = 0;
 static int verbose   = 0;
-static int names     = 0;
 const char *users[MAXNAMES];
 const char *hosts[MAXNAMES];
 int command=SHOW,num_users=0, num_hosts=0;
@@ -155,28 +153,36 @@ static void showblocking(const abl_args *args, const char *rule, const time_t *h
     }
 }
 
-static int doshow(const abl_args *args, const char *rule, const char *dbfile, const char *thing) {
+static int doshow(const abl_args *args, abl_info *info) {
     DB *db=NULL;
-    int err, err2;
     DBT key, data;
     DBC *cursor;
-    time_t now = time(NULL);
-    char *buf = NULL;
+    int err;
     int bsz = 0;
     int cnt = 0;
+    char *buf = NULL, *thing = NULL;
+    const char *rule;
+    time_t now = time(NULL);
 
-    if (NULL == dbfile) {
+    if (args == NULL || info== NULL) {
         return 0;
     }
 
+    if( info->subject == HOST ) {
+        db = info->htdb;
+        thing = "hosts";
+        rule = args->host_rule;
+    }
+    else {
+        db = info->utdb;
+        thing = "users";
+        rule = args->user_rule;
+    }
+
+
     //Section header for output
     printf("Failed %s:\n", thing);
 
-    err = dbopen(args, dbfile, DB_TIME, &db);
-    if (err != 0) {
-        goto doshow_fail;
-    }
-
     memset(&key,  0, sizeof(key));
     memset(&data, 0, sizeof(data));
 
@@ -211,7 +217,7 @@ static int doshow(const abl_args *args, const char *rule, const char *dbfile, co
 
         memcpy(buf, key.data, key.size);
         buf[key.size] = '\0';
-        printf(PAD "%s (%d)\n", buf, data.size / sizeof(time_t));
+        printf(PAD "%s (%ld)\n", buf, data.size / sizeof(time_t));
         cnt++;
 
         tm = (time_t *) data.data;
@@ -237,39 +243,42 @@ static int doshow(const abl_args *args, const char *rule, const char *dbfile, co
     /* Cleanup */
 doshow_fail:
     free(buf);
-
-    if (err2 = db->close(db, 0), 0 != err2) {
-        log_sys_error(err2, "closing database");
-        if (0 == err) {
-            err = err2;
-        }
-    }
-
     return err;
 }
 
-static int dopurge(const abl_args *args, const char *dbfile, long maxage) {
-    DB *db = NULL;
-    int err, err2;
+static int dopurge(const abl_args *args, abl_info *info) {
+    DB *tdb = NULL;
+    DB *sdb = NULL;
     DBT key, data;
     DBC *cursor;
+    int err;
+    long maxage = 0;
+    const char **thing;
     time_t now = time(NULL);
 
-    if (NULL == dbfile) {
+    if ( args == NULL || info == NULL ) {
         return 0;
     }
 
-    err = dbopen(args, dbfile, DB_TIME, &db);
-    if (err != 0) {
-        goto dopurge_fail;
+    if(info->subject == HOST) {
+        tdb = info->htdb;
+        sdb = info->hsdb;
+        maxage = args->host_purge;
+        thing = &info->host;
     }
-
-    mention("Purging %s", dbfile);
+    else {
+        tdb = info->utdb;
+        sdb = info->usdb;
+        maxage = args->user_purge;
+        thing = &info->user;
+    }
+    /*Cheat, because get_dbname is broken starting on some version lower than 4.8 */
+    mention("Purging %s", tdb->fname);
 
     memset(&key,  0, sizeof(key));
     memset(&data, 0, sizeof(data));
 
-    if (err = db->cursor(db, NULL, &cursor, 0), 0 != err) {
+    if (err = tdb->cursor(tdb, NULL, &cursor, 0), 0 != err) {
         log_sys_error(err, "creating cursor");
         goto dopurge_fail;
     }
@@ -285,9 +294,15 @@ static int dopurge(const abl_args *args, const char *dbfile, long maxage) {
 
         if (rule_purge(&data, maxage, now)) {
             if (data.size == 0) {
-                if (err = cursor->c_del(cursor, 0), 0 != err) {
+                err = ( cursor->c_del(cursor, 0) || sdb->del(sdb, NULL, &key, 0) ); 
+                if (err) {
                     goto dopurge_fail;
                 }
+                info->state = CLEAR;
+                *thing = (char*) key.data;
+                update_status(args, info);
+                if (info->state_change) run_command(args,info);
+
             } else {
                 if (err = cursor->c_put(cursor, &key, &data, DB_CURRENT), 0 != err) {
                     goto dopurge_fail;
@@ -298,37 +313,40 @@ static int dopurge(const abl_args *args, const char *dbfile, long maxage) {
 
     /* Cleanup */
 dopurge_fail:
+#if DB_VERSION_MAJOR < 5
     if (cursor != NULL) 
-#if DB_VERSION_MINOR <= 5
         cursor->c_close;
 #else
+    if (cursor != NULL) 
         cursor->close(cursor);
 #endif
 
-    err2 = db->close(db,0);
-    if(err2 != 0) {
-        log_sys_error(err2, "closing database");
-        if (0 == err) {
-            err = err2;
-        }
-    }
-
     return err;
 }
 
-static int whitelist(const abl_args *args, const char *dbfile, const char **permit, int count) {
+static int whitelist(const abl_args *args, abl_info *info, const char **permit, int count) {
     DB *db;
-    int err, err2;
+    int err = 0;
     DBT key, data;
     DBC *cursor;
     int wild = 0;
     int n, hit, del = 0;
     int *done;
+    const char **thing;
 
-    if (NULL == dbfile) {
+    if (args == NULL || info == NULL) {
         return 0;
     }
 
+    if(info->subject == HOST) {
+        db = info->htdb;
+        thing = &info->host;
+    }
+    else {
+        db = info->utdb;
+        thing = &info->user;
+    }
+
     if (done = malloc(sizeof(int) * count), NULL == done) {
         log_sys_error(ENOMEM, "allocating buffer");
     }
@@ -342,11 +360,6 @@ static int whitelist(const abl_args *args, const char *dbfile, const char **perm
         done[n] = 0;
     }
 
-    err = dbopen(args, dbfile, DB_TIME, &db);
-    if (err != 0) {
-        goto whitelist_fail;
-    }
-
     memset(&key,  0, sizeof(key));
     //key.flags = DB_DBT_REALLOC;
     memset(&data, 0, sizeof(data));
@@ -380,6 +393,10 @@ static int whitelist(const abl_args *args, const char *dbfile, const char **perm
                 if (err = cursor->c_del(cursor, 0), 0 != err) {
                     goto whitelist_fail;
                 }
+                info->state = CLEAR;
+                *thing = permit[n];
+                update_status(args, info);
+                if (info->state_change) run_command(args,info);
                 del++;
             }
         }
@@ -393,6 +410,10 @@ static int whitelist(const abl_args *args, const char *dbfile, const char **perm
                     int m;
                     done[n] = 1;
                     del++;
+                    info->state = CLEAR;
+                    *thing = permit[n];
+                    update_status(args, info);
+                    if (info->state_change) run_command(args,info);
                     /* Slightly pedantic - knock out any duplicates so we don't flag an error when
                      * we can't delete them later. Arguably if you supply duplicate arguments you
                      * should expect an error but this keeps the semantics the same as the wildcard
@@ -432,23 +453,15 @@ static int whitelist(const abl_args *args, const char *dbfile, const char **perm
 whitelist_fail:
     free(done);
 
-    if (err2 = db->close(db, 0), 0 != err2) {
-        log_sys_error(err2, "closing database");
-        if (0 == err) {
-            err = err2;
-        }
-    }
-
     return err;
 }
 
-static int fail(const abl_args *args, const char *dbfile, const char **prohibitions, int count, int max_age) {
-    int n, err, add = 0;
-    int wild = 0;
+static int fail(const abl_args *args, abl_info *info, const char **prohibitions, int count, int max_age) {
+    int n, err = 0, add = 0;
     int *done;
     time_t tm = time(NULL);
 
-    if (NULL == dbfile) {
+    if (args == NULL || info == NULL) {
         return 0;
     }
 
@@ -468,7 +481,16 @@ static int fail(const abl_args *args, const char *dbfile, const char **prohibiti
     /* add individual items */
     for (n = 0; n < count; n++) {
         if (!done[n]) {
-            err = record(args,dbfile,prohibitions[n],tm,max_age);
+            if(info->subject == HOST) {
+                info->host = prohibitions[n];
+                err = record(args,info,tm,max_age);
+                if(!err) check_host(args,info,tm);
+            } else {
+                info->user = prohibitions[n];
+                err = record(args,info,tm,max_age);
+                if(!err) check_user(args,info,tm);
+            }
+
             if (0 == err) {
                 int m;
                 done[n] = 1;
@@ -494,33 +516,36 @@ static int fail(const abl_args *args, const char *dbfile, const char **prohibiti
     }
 
     /* Cleanup */
-fail:
     free(done);
 
     return err;
 }
 
 int main(int argc, char **argv) {
-    int argn, err, n, c, rv;
-    int digit_optind = 0;
+    int err, n, c;
     char *conf = NULL, *service = "none";
     abl_args args;
+    abl_info info;
+    DB *utdb;
+    DB *usdb;
+    DB *htdb;
+    DB *hsdb;
 
     while (1) {
-        int this_option_optind = optind ? optind : 1;
+
         int option_index = 0;
         static struct option long_options[] = {
-            {"help",        0, &help,       1        },
-            {"purge",       0, &command,    PURGE    },
-            {"relative",    0, &relative,   1        },
-            {"verbose",     0, &verbose,    1        },
-            {"fail",        1, &command,    FAIL     },
-            {"whitelist",   1, &command,    WHITELIST},
-            {"check",       1, &command,    CHECK    },
-            {"user",        1, 0,           'U'      },
-            {"host",        1, 0,           'H'      },
-            {"service",     1, 0,           's'      },
-            {0,             0, 0,           0        }
+            {"help",        0, 0,    'h' },
+            {"purge",       0, 0,    'p' },
+            {"relative",    0, 0,    'r' },
+            {"verbose",     0, 0,    'v' },
+            {"fail",        1, 0,    'f' },
+            {"whitelist",   1, 0,    'w' },
+            {"check",       1, 0,    'c' },
+            {"user",        1, 0,    'U' },
+            {"host",        1, 0,    'H' },
+            {"service",     1, 0,    's' },
+            {0,             0, 0,     0  }
         };
 
         c = getopt_long(argc, argv, "hrvpfwcU:H:s:",
@@ -562,7 +587,7 @@ int main(int argc, char **argv) {
                 break;
 
             default:
-                printf("Unkown error parsing command line arguments: code 0%o\n", c);
+                printf("Unknown error parsing command line arguments: code 0%o\n", c);
         }
     }
 
@@ -597,13 +622,33 @@ int main(int argc, char **argv) {
     for(n=0;n<num_hosts;n++){
         log_debug(&args,"host: %s",hosts[n]);
     }
+    memset(&info, 0, sizeof(abl_info));
+
+    /* Most everything should be set, and it should be safe to open the
+     * databases. */
+    err = dbopen(&args, args.user_db, DB_TIME, &utdb);
+    if (err) return err;
+    info.utdb = utdb;
+    err = dbopen(&args, args.user_db, DB_STATE, &usdb);
+    if (err) return err;
+    info.usdb = usdb;
+    err = dbopen(&args, args.host_db, DB_TIME, &htdb);
+    if (err) return err;
+    info.htdb = htdb;
+    err = dbopen(&args, args.host_db, DB_STATE, &hsdb);
+    if (err) return err;
+    info.hsdb = hsdb;
+
+
 
     if (command == WHITELIST) {
         if (num_users > 0) {
-            whitelist(&args, args.user_db, users, num_users);
+            info.subject = USER;
+            whitelist(&args, &info, users, num_users);
         }
         if (num_hosts > 0) {
-            whitelist(&args, args.host_db, hosts, num_hosts);
+            info.subject = HOST;
+            whitelist(&args, &info, hosts, num_hosts);
         }
         if (num_users == 0 && num_hosts == 0) {
             die("Asked to whitelist but no hosts or users given!");
@@ -611,10 +656,12 @@ int main(int argc, char **argv) {
     }
     else if (command == FAIL) {
         if (num_users > 0) {
-            fail(&args, args.user_db, users, num_users, args.user_purge);
+            info.subject = USER;
+            fail(&args, &info, users, num_users, args.user_purge);
         }
         if (num_hosts > 0) {
-            fail(&args, args.host_db, hosts, num_hosts, args.host_purge);
+            info.subject = HOST;
+            fail(&args, &info, hosts, num_hosts, args.host_purge);
         }
         if (num_users == 0 && num_hosts == 0) {
             die("Asked to fail someone, but no hosts or users given!");
@@ -624,8 +671,6 @@ int main(int argc, char **argv) {
     if (command == CHECK) {
         int  err = 0;
         time_t tm = time(NULL);
-        abl_info info;
-        memset(&info,0,sizeof(abl_info));
 
         if (num_users == 0 && num_hosts == 0) {
             die("Asked to check but no hosts or users given!");
@@ -641,28 +686,24 @@ int main(int argc, char **argv) {
 
         if (num_hosts > 0) {
             info.host = hosts[0];
-            info.dbfile = args.host_db;
+            info.subject = HOST;
             err = check_host(&args, &info, tm);
             if(err) {
                 log_warning(&args,"Could not check host.");
             }
             else {
-                update_status(&args, &info);
-                if (info.state_change) run_command(&args,&info);
                 if (info.state == BLOCKED) exit(info.state);
             }
         }
 
         if (num_users > 0) {
             info.user = users[0];
-            info.dbfile = args.user_db;
+            info.subject = USER;
             err = check_user(&args, &info, tm);
             if(err) {
                 log_warning(&args,"Could not check user.");
             }
             else {
-                update_status(&args, &info);
-                if (info.state_change) run_command(&args,&info);
                 if (info.state == BLOCKED) exit(info.state);
             }
         }
@@ -671,15 +712,24 @@ int main(int argc, char **argv) {
 
 
     if (command == PURGE) {
-        dopurge(&args, args.user_db, args.user_purge);
-        dopurge(&args, args.host_db, args.host_purge);
+        info.subject = USER;
+        dopurge(&args, &info);
+        info.subject = HOST;
+        dopurge(&args, &info);
     } else if (num_users == 0 && num_hosts == 0) {
-        doshow(&args, args.user_rule, args.user_db, "users");
-        doshow(&args, args.host_rule, args.host_db, "hosts");
+        info.subject = USER;
+        doshow(&args, &info);
+        info.subject = HOST;
+        doshow(&args, &info);
     }
 
     config_free(&args);
 
+    err = utdb->close(utdb,0);
+    err |= usdb->close(usdb,0);
+    err |= htdb->close(htdb,0);
+    err |= hsdb->close(hsdb,0);
+
     return 0;
 }
 
